@@ -59,9 +59,8 @@
 
 ;; Other open problems are `setenv'/`getenv'.
 
-;; If `start-file-process' isn't defined, it gets an alias in tramp-compat.el.
-(when (eq (symbol-function 'start-file-process)
-	  'tramp-handle-start-file-process)
+(unless (tramp-exists-file-name-handler 'start-file-process "/")
+
   (defadvice executable-find
     (around tramp-advice-executable-find activate)
     "Invoke `tramp-handle-executable-find' for Tramp files."
@@ -106,7 +105,9 @@
 	    ad-do-it))
       ad-do-it))
   (add-hook 'tramp-util-unload-hook
-	    '(lambda () (ad-unadvise 'start-process-shell-command)))
+	    '(lambda () (ad-unadvise 'start-process-shell-command))))
+
+(unless (tramp-exists-file-name-handler 'process-file "/")
 
   (defadvice call-process
     (around tramp-advice-process-file activate)
@@ -162,28 +163,6 @@
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-;; In Emacs 21, `dired-insert-directory' calls
-;; `dired-free-space-program'.  This isn't aware of remote
-;; directories.  It will be disabled temporarily.
-
-(when (and (boundp 'dired-free-space-program)
-	   (not (boundp 'directory-free-space-program))) ;; Alias in Emacs 22.
-  (defadvice dired-insert-directory
-    ;; Don't use ARGS from `dired-insert-directory', they are
-    ;; different for (X)Emacs flavors.
-    (around tramp-advice-dired-insert-directory activate)
-    "Disable `dired-free-space-program' for Tramp files."
-    (let ((dired-free-space-program
-	   (and (not
-		 (eq (tramp-find-foreign-file-name-handler default-directory)
-		     'tramp-sh-file-name-handler))
-		(symbol-value 'dired-free-space-program))))
-      ad-do-it))
-  (add-hook 'tramp-util-unload-hook
-	    '(lambda () (ad-unadvise 'dired-insert-directory))))
-
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-
 ;; compile.el parses the compilation output for file names.  It
 ;; expects them on the local machine.  This must be changed.
 
@@ -314,7 +293,7 @@ If it is an absolute file name, and not a remote one, prepend the remote part."
     ;; Emacs 22 uses `gud-file-name' which we should do as well.
     ;; `gud-<MINOR-MODE>-directories' must be Tramp file names.
     (if (functionp 'gud-file-name)
-	(funcall (symbol-function 'gud-file-name) filename)
+	(tramp-compat-funcall 'gud-file-name filename)
       filename)))
 
 (defun tramp-gud-massage-args (args)
@@ -377,11 +356,9 @@ Works only for relative file names and Tramp file names."
 
  ;; So far, I've tested only gdb and perldb.
  ;; (X)Emacs
- '(gdb sdb dbx xdb perldb
+ '(gdb sdb dbx xdb perldb pdb
  ;; Emacs
-   pdb jdb
- ;; Emacs 22
-   bashdb))
+   jdb bashdb))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
